#include <stdlib.h>
#include <iconv.h>
#include <langinfo.h>
#include <errno.h>

#include <X11/Xlib.h>
#include <X11/cursorfont.h>

#include "tategaki.h"
#include "t_screen.h"

char lastchar(const char *str)
{
	return str[strlen(str) - 1];
}


static char *codeset[] = { "EUC-JP", "SHIFT_JIS", "UTF-8", "ISO-8859-1", NULL };

int is_encoding(char *input, char *encoding)
{
	size_t inputsize = strlen(input) + 1;
	size_t bufsize = inputsize*MB_CUR_MAX;
	char *buf_start = NULL;
	char *str, *buf;
	size_t res;
	iconv_t cd;


	buf_start = calloc(1, bufsize);

	str = input, buf = buf_start;

	cd = iconv_open("EUC-JP", encoding);
	if (cd == (iconv_t) -1) {
		perror("iconv_open");
	}
	
	res = iconv(cd, &str, &inputsize, &buf, &bufsize);
//		DEBUG_INT(E2BIG == errno);
//		DEBUG_INT(EILSEQ == errno);
//		DEBUG_INT(EINVAL == errno);
	iconv_close(cd);

	free(buf_start);

	return (res == -1) ? 0 : 1;
}

char *determine_codeset(char *input)
{
	int i;
	static int guess = 0; /* index to codeset[] */

	if (is_encoding(input, codeset[guess])) {
		return codeset[guess];
	}

	for (i = 0; codeset[i] != NULL; i++) {
		if (i == guess)
			continue;

		if (is_encoding(input, codeset[i])) {
			return codeset[i];
		}
	}

	return NULL;
}

void strchop(char *str)
{
	int end_pos = strlen(str) - 1;

	if (str[end_pos] == '\n' || str[end_pos] == '\r') {
		str[end_pos] = '\0';
		return strchop(str);
	} else {
		return;
	}
}

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

char *codeset_of_file(const char *filename)
{
	struct stat st;
	FILE *f;
	char *buf;
	char *code;

	if (stat(filename, &st) == -1) {
		perror("stat");
		exit(1);
	}

	buf = calloc(1, st.st_size);
	f = fopen(filename, "r");
	if (f == NULL) {
		perror("fopen");
		exit(1);
	}
	fread(buf, 1, st.st_size, f);
	code = determine_codeset(buf);
	free(buf);

	return code;
}

char *fgetline(FILE *f, const char *file_code)
{
	size_t size = 128;
	char *buf = NULL, *internal = NULL;
	iconv_t cd;
	size_t buf_size, internal_size;
	char *buf_copy, *internal_copy;
	size_t res;
	char tmp[128];

	if (feof(f))
		return NULL;
	buf = calloc(128, 1);

	while (fgets(tmp, 128, f) != NULL) {
		if (strlen(buf) + strlen(tmp) >= size) {
			size *= 2;
			buf = realloc(buf, size);
		}
		strcat(buf, tmp);
		if (lastchar(tmp) == '\n') {
			break;
		}
	}
	strchop(buf);

	cd = iconv_open(nl_langinfo(CODESET), file_code);
	if (cd == (iconv_t) -1)
		perror("iconv_open"), exit(1);

	buf_size = strlen(buf) + 1;
	internal_size = buf_size*MB_CUR_MAX;
	internal = realloc(internal, internal_size);
	buf_copy = buf; internal_copy = internal;

	do {
		res = iconv(cd, &buf_copy, &buf_size, &internal_copy, &internal_size);
		if (res == (size_t) -1) {
			if (errno == E2BIG) {
				internal_size *= 2;
				internal = realloc(internal, buf_size);
			} else {
				perror("iconv");
				exit(1);
			}
		}
	} while (res == -1);

	return internal;
}



static void t_screen_set_size(t_screen *self, int global_length, int line_length)
{
	self->screen_geo->chokkaku_len = global_length;
	self->screen_geo->shinkou_len = line_length;

	self->drawer->ext.width = global_length;
	self->drawer->ext.height = line_length;
}

static void t_screen_load_file(t_screen *self, const char *filename)
{
	FILE *fp;
	char *buf = NULL;
	wchar_t *wbuf = NULL;
	char *file_code;

	file_code = codeset_of_file(filename);

	fp = fopen(filename, "r");
	if (fp == NULL) {
		perror(filename);
		exit(1);
	}


	while ((buf = fgetline(fp, file_code))) {
		line *new;
		size_t nchars = mbstowcs(NULL, buf, 0);
		if (nchars == -1)
			DIE("mbstowcs returns -1.\n");

		wbuf = realloc(wbuf, nchars * sizeof(wchar_t) + 1);
		mbstowcs(wbuf, buf, nchars + 1);
		new = line_new(wbuf, self->drawer, self->screen_geo);
		t_screen_add_line(self, new);
	}

	return;
}

void t_screen_add_line(t_screen *self, line *new)
{
	self->lines[self->nlines] = new;
	self->nlines++;
}


t_screen *t_screen_new(const char *filename, int global_length, int line_length)
{
	t_screen *self = calloc(1, sizeof(t_screen));

	self->drawer = t_drawer_new(global_length, line_length, DEFAULT_FONT);
	self->geo = geometry_new();
	self->screen_geo = geometry_new();
	t_screen_set_size(self, global_length, line_length);

	self->lines = calloc(1024, sizeof(line *));
	self->nlines = 0;

	t_screen_load_file(self, filename);

	return self;
}


void t_screen_calc_geometry(t_screen *self, int shinkou)
{
	int i;
	geometry *origin = geometry_copy(self->geo);

	self->geo->chokkaku_len = 0;
	for (i = 0; i < self->nlines; i++) {
		line *l = self->lines[i];

		line_set_origin(l, origin);
		line_calc_geometry(l, self->screen_geo->shinkou_len);
		origin->chokkaku_pos += l->geo->chokkaku_len; /* ľ˿ʹ */

		self->geo->chokkaku_len += l->geo->chokkaku_len;
		if (self->geo->shinkou_len < l->geo->shinkou_len)
			self->geo->shinkou_len = l->geo->shinkou_len;
	}

	free(origin);
}

void t_screen_free(t_screen *self)
{
	/* äȤȤ.... ̤ */
}

#include <X11/keysym.h>

geometry *t_screen_xexpose_to_geometry(t_screen *self, XExposeEvent *e)
{
	geometry *rv = geometry_new();

	/* XXX depend orientation */

	rv->shinkou_pos = e->y;
	rv->chokkaku_pos = self->screen_geo->chokkaku_len - (e->x + e->width);
	rv->shinkou_len = e->height;
	rv->chokkaku_len = e->width;

	rv->shinkou_pos += self->screen_geo->shinkou_pos;
	rv->chokkaku_pos += self->screen_geo->chokkaku_pos;

	return rv;
}

void t_screen_main(t_screen *self)
{
	XEvent event;

	for (;;) {
		t_drawer_get_next_event(self->drawer, &event);
		switch(event.type){
		case Expose:
		{
			XExposeEvent *e = &event.xexpose;
			geometry *area_to_redraw;
			area_to_redraw = t_screen_xexpose_to_geometry(self, e);

//			DEBUG_PRINT("Expose\n");
			t_screen_expose_callback(self, area_to_redraw);
			break;
		}
                case ClientMessage:
                        if(event.xclient.data.l[0] == self->drawer->protocols){
                                t_drawer_destroy(self->drawer);
                        }
                        break;
		case ConfigureNotify:
		{
			XConfigureEvent *new = &event.xconfigure;

			if (new->width == self->screen_geo->chokkaku_len &&
			    new->height == self->screen_geo->shinkou_len)
				break; /* Moving. Not resizing.*/

//			DEBUG_PRINT("Resize\n");
			t_drawer_define_cursor(self->drawer, XC_watch);
			t_screen_set_size(self, new->width, new->height);
			t_screen_calc_geometry(self, self->screen_geo->shinkou_len);
			t_drawer_clear_area(self->drawer, 0,0,0,0, True);
			t_drawer_undefine_cursor(self->drawer);


			break;
		}
		case KeyPress:
		{
			KeySym keysym;
			char buffer[10];
			XLookupString(&event.xkey, buffer, sizeof buffer, &keysym, NULL);
			int line_chokkaku_scale = self->lines[self->focus_pos]->geo->chokkaku_len;

			switch (keysym) {
			case XK_Left:
				t_screen_scroll_left(self, line_chokkaku_scale);
				break;
			case XK_Right:
				t_screen_scroll_right(self, line_chokkaku_scale);
				break;
			case XK_Prior:
				t_screen_scroll_right(self, self->screen_geo->chokkaku_len - line_chokkaku_scale*2);
				break;
			case XK_Next:
				t_screen_scroll_left(self, self->screen_geo->chokkaku_len - line_chokkaku_scale*2);
				break;
			case XK_Home:
				t_screen_jump_to(self, 0);
				break;
			case XK_End:
				t_screen_jump_to(self, self->geo->chokkaku_len - self->screen_geo->chokkaku_len);
				break;
			default:
//				DEBUG_PRINT("default\n");
				break;
			}
		}
		case ButtonPress:
		{
			XButtonEvent *e = &event.xbutton;

			DEBUG_PRINT("ButtonPress\n");
			switch (e->button) {
			case Button1:
			case Button2:
				self->grabbed = 1;
				self->grab_x = e->x;
				self->grab_y = e->y;
				break;
			default:
				break;
			}
			break;
		}
		case ButtonRelease:
		{
			XButtonEvent *e = &event.xbutton;

			DEBUG_PRINT("ButtonRelease\n");
			switch (e->button) {
			case Button1:
			case Button2:
				self->grabbed = 0;
			default:
				break;
			}
			break;
		}
		case MotionNotify:
		{
			XMotionEvent *e = &event.xmotion;

			DEBUG_PRINT("MotionNotify\n");
			DEBUG_INT(self->grabbed);
			if (self->grabbed) {
				int distance = self->grab_x - e->x;

				DEBUG_INT(distance);
				if (distance > 0) {
					t_screen_scroll_right(self, distance);
				} else if (distance < 0) {
					t_screen_scroll_left(self, - distance);
				}
				self->grab_x = e->x;
				self->grab_y = e->y;
			}
			break;
		}
		default:
			break;
		}
	}
	
	return;
}

void t_screen_expose_callback(t_screen *self, geometry *area_to_redraw)
{
	int i;

	for (i = 0; i < self->nlines; i++) {
		line *l = self->lines[i];

		if (geometry_overlapped_p(l->geo, area_to_redraw)) {
			line_expose_callback(l, area_to_redraw);
		}
	}
}


void t_screen_scroll_left(t_screen *self, int width)
{
	int max_chokkaku_pos = self->geo->chokkaku_len - self->screen_geo->chokkaku_len;


	if (self->screen_geo->chokkaku_pos >= max_chokkaku_pos)
		return;

	if (self->screen_geo->chokkaku_pos + width > max_chokkaku_pos) {
		width = max_chokkaku_pos - self->screen_geo->chokkaku_pos;
	}
	self->screen_geo->chokkaku_pos += width;
	if (self->screen_geo->chokkaku_pos < 0)
		self->screen_geo->chokkaku_pos = 0;

	t_drawer_copy_area(self->drawer,
			   0, 0,
			   self->drawer->ext.width - width, self->drawer->ext.height,
			   0 + width, 0);

	t_drawer_clear_area(self->drawer,
			    0, 0,
			    width, self->drawer->ext.height,
			    True);


	return;
}

void t_screen_scroll_right(t_screen *self, int width)
{
	if (self->screen_geo->chokkaku_pos <= 0)
		return ;

	if (self->screen_geo->chokkaku_pos - width < 0) {
		width = self->screen_geo->chokkaku_pos;
	}
	self->screen_geo->chokkaku_pos -= width;

	t_drawer_copy_area(self->drawer,
			   0 + width, 0,
			   self->drawer->ext.width, self->drawer->ext.height,
			   0, 0);
	t_drawer_clear_area(self->drawer,
			    self->drawer->ext.width - width, 0,
			    width, self->drawer->ext.height,
			    True);

	return;
}

void t_screen_jump_to(t_screen *self, int chokkaku_pos)
{
	self->screen_geo->chokkaku_pos = chokkaku_pos;
	/* ExposeEvent */
	t_drawer_clear_area(self->drawer, 0,0,0,0, True);
}
